#define _USE_BSD
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>

#define RETSIGTYPE void
/* A Pointer to a signal handler.  */
typedef RETSIGTYPE (*sighandler) ();

void run_command (char *const *cmd)
{
  pid_t pid;			/* Pid of child.  */
  pid_t my_pid;
  sighandler interrupt_signal, quit_signal;

  my_pid = getpid();

  pid = fork ();		/* Run CMD as child process.  */
  if (pid < 0)
    {
      fprintf(stderr,"cannot fork.\n");
      exit(-1);
    }
  else if (pid == 0)
    {				/* If child.  */
      /* Don't cast execvp arguments; that causes errors on some systems,
	 versus merely warnings if the cast is left off.  */
      execvp (cmd[0], cmd);
      fprintf(stderr,"pperf: Cannot run %s\n\t",cmd[0]);
      perror("");
      kill(my_pid,SIGKILL);
      exit(-1);
    }

  /* Have signals kill the child but not self (if possible).  */
  interrupt_signal = signal (SIGINT, SIG_IGN);
  quit_signal = signal (SIGQUIT, SIG_IGN);

  if (wait_for_job (pid) == 0)
    {
      fprintf(stderr,"error waiting for child process.\n");
      exit(-1);
    }

  /* Re-enable signals.  */
  signal (SIGINT, interrupt_signal);
  signal (SIGQUIT, quit_signal);
}

int wait_for_job (pid_t pid)
{
  int status;

  pid_t caught;

  /* Ignore signals, but don't ignore the children.  When wait3
     returns the child process, set the time the command finished. */

  while ((caught = wait3 (&status, 0, NULL)) != pid)
    {
      if (caught == -1)
	return 0;
    }

}

