/*
 * Machine Specific Register Driver for Linux
 * by Stephan Meyer
 */

/* need to do "mknod /dev/msr c 10 142" to make the device */

#include <linux/config.h>
#include <linux/module.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <linux/types.h>
#include <linux/stddef.h>
#include <linux/timer.h>
#include <linux/fcntl.h>
#include <linux/malloc.h>
#include <linux/linkage.h>
#include <linux/miscdevice.h>

#define MSR_MINOR 142

/*
 * FIXME: why does misc.c require nonNULL open()/close() functions?
 */
  
static int msr_open(struct inode * inode, struct file * file)
{
	return 0;
}

static void msr_close(struct inode * inode, struct file * file)
{
}

static long msr_read(struct inode *inode, struct file *file,
                               char *buf, unsigned long count)
{
	unsigned long int p=file->f_pos,hi,lo;
	__asm__ __volatile__ (
		"rdmsr"
		:"=a" (lo), "=d" (hi)
		:"c" (p)
		:"eax","ecx","edx");

        /*
         * We are x86 only, and we absolutely assume the above is 64 bit data
         * put into 32 bit registers.
         */                                   
                                   
	if (copy_to_user(buf,&hi,4))
		return -EFAULT;
	if (copy_to_user(buf+4,&lo,4))
		return -EFAULT;
	return 8;
}

static long msr_write(struct inode *inode, struct file *file,
                               const char *buf, unsigned long count)
{
	unsigned long p=file->f_pos,hi,lo;
	
	if (copy_from_user(&hi,buf,4))
		return -EFAULT;
	if (copy_from_user(&lo,buf+4,4))
		return -EFAULT;
	
	__asm__ __volatile__ (
		"wrmsr"
		:
		:"c" (p), "a" (lo), "d" (hi)
		:"eax","ecx","edx");
	return 8;
}

static long long msr_seek(struct inode *inode, struct file *file,
                             long long offset, int orig)
{
	switch (orig) {
		case 0:	
			file->f_pos=offset;
			return file->f_pos;
		case 1:	
			file->f_pos+=offset;
			return file->f_pos;
		default:
			return -EINVAL;
	}
}

static struct file_operations msr_fops = {
	msr_seek,
	msr_read,
	msr_write,
	NULL,
	NULL,
	NULL,
	NULL,
	msr_open,
	msr_close,
	NULL,
	NULL
};

static struct miscdevice msr_device = {
	MSR_MINOR,
	"msr",
	&msr_fops
};

#ifndef MODULE	
void msr_init(void)
{
	if ((!have_cpuid) || (!(x86_capability & 32))) {
		printk("MSR not supported!\n");
		return;
	}
	printk("MSR device installed.\n");
	
	misc_register(&msr_device);
}
#endif

#ifdef MODULE

int init_module(void)
{
        unsigned int cpu;

        asm("cpuid\n\t"
	    : "=a" (cpu)
	    : "a" (0)
	    : "%eax", "%ebx", "%ecx", "%edx");

	if (cpu != 1) /* Not Pentium */
	  {
	    printk(KERN_INFO "\nEnabling user-level RDPMC.\n");
	    /* Enable RDPMC at any protection level */
	    asm("movl %cr4,%eax\n"); 
	    asm("orl $256,%eax\n");
	    asm("movl %eax,%cr4\n");
	  }

        misc_register(&msr_device);

        printk(KERN_INFO "MSR device installed.\n");

	return 0;
}

void cleanup_module(void)
{
        misc_deregister(&msr_device);
}
#endif /* MODULE */


/*
 * Local variables:
 *  compile-command: "gcc -DMODULE -D__KERNEL__ -I/usr/src/linux/include -Wall -Wstrict-prototypes -O2 -c msr.c"
 * End:
 */
