/*
 * This code is from lmbench with original id string:
 * char    *id = "$Id: speed.c,v 1.1.1.1 1998/10/07 08:09:02 zdv057 Exp $\n";
 * This code originally appeared in mhz.c and is covered as below:
 * Copyright (c) 1994 Larry McVoy.  Distributed under the FSF GPL.
 *
 * Modifications of this code where made by M. Patrick Goda, 1997
 * for inclusion into pperf.  
 *
 * mpg 07 Nov 97
 */

#include <sys/time.h>

static struct timeval start_tv, stop_tv;

#ifdef  RUSAGE
#include <sys/resource.h>
#include <stdio.h>
#define secs(tv)        (tv.tv_sec + tv.tv_usec / 1000000.0)
#define mine(f)         (ru_stop.f - ru_start.f)

static struct rusage ru_start, ru_stop;

void
rusage()
{
        struct rusage r;
        double  sys, user, idle;
        double  per;
        double  timespent();

        sys = secs(ru_stop.ru_stime) - secs(ru_start.ru_stime);
        user = secs(ru_stop.ru_utime) - secs(ru_start.ru_utime);
        idle = timespent() - (sys + user);
        per = idle / timespent() * 100;
        fprintf(stderr, "real=%.2f sys=%.2f user=%.2f idle=%.2f stall=%.0f%% ",
            timespent(), sys, user, idle, per);
        fprintf(stderr, "rd=%d wr=%d min=%d maj=%d ctx=%d\n",
            mine(ru_inblock), mine(ru_oublock),
            mine(ru_minflt), mine(ru_majflt),
            mine(ru_nvcsw) + mine(ru_nivcsw));
}

#endif

void
tvsub(tdiff, t1, t0)
        struct timeval *tdiff, *t1, *t0;
{
        tdiff->tv_sec = t1->tv_sec - t0->tv_sec;
        tdiff->tv_usec = t1->tv_usec - t0->tv_usec;
        if (tdiff->tv_usec < 0)
                tdiff->tv_sec--, tdiff->tv_usec += 1000000;
}

#ifdef RUSAGE
double
timespent()
{
        struct timeval td;

        tvsub(&td, &stop_tv, &start_tv);
        return (td.tv_sec + td.tv_usec / 1000000);
}
#endif

ulong
stop()
{
        struct timeval tdiff;

        (void) gettimeofday(&stop_tv, (struct timezone *) 0);
#ifdef  RUSAGE
        getrusage(RUSAGE_SELF, &ru_stop);
#endif

        tvsub(&tdiff, &stop_tv, &start_tv);
        return (tdiff.tv_sec * 1000000 + tdiff.tv_usec);
}

void
start()
{
        (void) gettimeofday(&start_tv, (struct timezone *) 0);
#ifdef  RUSAGE
        getrusage(RUSAGE_SELF, &ru_start);
#endif
}

#ifndef N
#define N       20000000
#endif

float speed( void )
{
	register a, i;
	int	result = 1024*1024*1024;
	int	tries = 0;
	double	mhz;


        for (tries = 0; tries < 5; ++tries) {
                /* time empty loop */
		i = N - 1000;
		a = 1024;
                start();
                while (i > 0) {
#define	FOUR	a++; a++; a++; a++;
#define TWENTY	FOUR FOUR FOUR FOUR FOUR
#define	H	TWENTY TWENTY TWENTY TWENTY TWENTY
			H H H H H
			H H H H H
			i -= 1000;
                }
		i = stop();
		if (i < result)
			result = i;
	}
	/*
	 * We figure that the loop overhead is about 5 
	 * instructions out of 1000.  So we figure that
	 * we actually execute N * 1.005 instructions.
	 */
	mhz = (N * 1.005) / result;
	return( mhz );
}


